#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: windfic@qq.com
# @Last Modified Time: 2017-03-17 20:36

import os
import re
import sys
import ast
import unparse
from converter import Processor
import xml.etree.ElementTree as Tree

class Compiler(Processor):

    def __init__(self, info=0):
        self.info = info
        self.xml = Tree.Element('root')
        self.handles = Handler('root')
        self.current = None
        self.visitor = None
        self.keywords = []
        self.state_machine = {}
        self.action_table = {}
        self.init_keywords()
        self.init_visitor()
        self.init_state_machine()
        self.init_action_table()

    def init_keywords(self):
        pass

    def init_state_machine(self):
        pass

    def init_action_table(self):
        pass

    def init_visitor(self):
        pass

    def set_path(self, root, app):
        self.visitor.set_path(root, app)

    def init_handles(self, src):
        h = Handler('file')
        h.set('value', src)
        self.handles.append(h)
        self.current = h

    def compile_file(self, src, src_enc='auto'):
        if not os.path.isfile(src):
            return False
        self.log('Parse %s' % src)

        if src_enc == 'auto':
            src_enc = self.detect(src)

        self.init_handles(src)

        # 语法解析开始
        self.parse_token('$$', '')

        sfh = open(src, 'r')
        try:
            for line in sfh.readlines():
                self.state_machine['current']['line'] += 1
                line = line.decode(src_enc)
                while True:
                    line = self.next_token(line)
                    if line is None or line is '':
                        break
        finally:
            sfh.close()

        # 语法解析结束
        if self.parse_token('$$', ''):
            self.log('Done!', 'success')
        else:
            self.log('Error!', 'error')

        self.init_state_machine()
        self.init_action_table()

        self.visitor.visit(self.current)

        return self

    # Tokenizer
    def next_token(self, line):
        state = self.state_machine['current']['state']
        match = None
        for n in self.state_machine[state]:
            if n['token'] is '':
                if match is None:
                    n['cache'] = line
                    n['start'] = 0
                    n['end'] = len(line)
                    match = n
                elif match['start'] > 0:
                    n['cache'] = line[:match['start']]
                    n['start'] = 0
                    n['end'] = len(n['cache'])
                    match = n
            else:
                if n['token'] == '[dynamic]':
                    n['token'] = '^' + line.strip()
                    n['dynamic'] = True
                    line = line[len(n['token']):]
                    if self.info > 0:
                        self.log('[dynamic]='+n['token'])
                m = re.search(n['token'], line)
                if m and (match is None or match['start'] > m.start() or
                          (match['start'] == m.start() and match['end'] < m.end())):
                    n['cache'] = m.group(0)
                    n['start'] = m.start()
                    n['end'] = m.end()
                    match = n
        if match is not None:
            self.state_machine['current']['content'] += match['cache']
            if match['next'] is not '':
                self.state_machine['current']['state'] = match['next']
                if 'dynamic' in match:
                    del match['dynamic']
                    match['token'] = '[dynamic]'
            if match['extra'] == 0:
                self.parse_token(match['name'], self.state_machine['current']['content'])
                self.state_machine['current']['content'] = ''
            match['cache'] = ''
            if match['end'] < len(line):
                return self.next_token(line[match['end']:])
        return None

    def get_keyword(self, name, token):
        if name == 'label':
            if token.lower() in self.keywords:
                return token.lower(), token
        elif name == 'symbol':
            return token, token
        return name, token

    # Parser
    def parse_token(self, name, token):
        (name, token) = self.get_keyword(name, token)
        if self.info > 2:
            self.log("%s:%s" % (name, token))
        self.action_table['stack'].append((name, token))

        while len(self.action_table['stack']) + len(self.action_table['extra']) >= 3:
            reduced = False
            name = self.action_table['stack'][-3]
            para = self.action_table['stack'][-2]
            extra = self.action_table['stack'][-1]
            if name[0] in self.action_table:
                if para[0] in self.action_table[name[0]]:
                    if extra[0] in self.action_table[name[0]][para[0]]:
                        if 'reduce' in self.action_table[name[0]][para[0]][extra[0]]:
                            value = self.action_table[name[0]][para[0]][extra[0]]['reduce']
                            operate = self.action_table[name[0]][para[0]][extra[0]]['op']
                            reduced = self.reduce(value, operate, 3)
                    else:
                        if 'reduce' in self.action_table[name[0]][para[0]]:
                            value = self.action_table[name[0]][para[0]]['reduce']
                            operate = self.action_table[name[0]][para[0]]['op']
                            reduced = self.reduce(value, operate, 2)
                else:
                    if 'reduce' in self.action_table[name[0]]:
                        value = self.action_table[name[0]]['reduce']
                        operate = self.action_table[name[0]]['op']
                        reduced = self.reduce(value, operate, 1)
            if not reduced:
                if len(self.action_table['extra']) == 0:
                    break
                else:
                    self.action_table['stack'].append(self.action_table['extra'].pop())
            while len(self.action_table['stack']) < 3 and len(self.action_table['extra']) > 0:
                self.action_table['stack'].append(self.action_table['extra'].pop())

        if len(self.action_table['stack']) < 3:
            if self.action_table['stack'][-1][0] == '$$':
                return True
            return False

    def reduce(self, state, operate, count):
        if state is '':
            return False  # 为空代表放弃本次reduce，因为前方有强匹配

        h = self.handle(state, operate, count)

        extra = self.action_table['stack'].pop()
        para = self.action_table['stack'].pop()
        if len(self.action_table['stack']) > 1:
            self.action_table['stack'].pop()

        if state != 'accept':
            self.action_table['stack'].append((state, h))
        if count < 3:
            self.action_table['extra'].append(extra)
        if count < 2:
            self.action_table['extra'].append(para)

        if self.info > 1:
            self.log("reduce:%s==>%s" % (h.content, state))

        return True

    def handle(self, state, operate, count):
        h = None
        if operate not in ('accept', 'ignore'):
            h = Handler(state, '')
        for i in range(-3, count - 3):
            old = self.action_table['stack'][i][0]
            if old == '$$':
                continue
            val = self.action_table['stack'][i][1]
            if not isinstance(val, Handler):
                val = Handler(old, val)
            if h is None:
                h = val
            else:
                if operate == 'merge':
                    if val.get('value') is '':
                        for c in val:
                            h.add(c, operate)
                    else:
                        h.add(val, operate)
                elif operate == 'prefix':
                    if val.get('value') is '' and i == -3:
                        for c in val:
                            h.add(c, operate)
                    else:
                        h.add(val, operate)
                elif operate == 'append':
                    if i > -3:
                        h[0].add(val, operate)
                    else:
                        h.add(val, operate)
                else:
                    h.add(val, operate)
        if state == 'accept':
            self.accept(state, operate, h)
        return h

    def accept(self, state, operate, val):
        if operate == 'ignore':
            return
        if isinstance(val, Handler):
            self.current.append(val)
            val = val.content
        if self.info > 0:
            self.log("%s:%s" % (state, val))

    # Executor
    def execute(self, handle=None):
        self.log('Execute ...')
        if handle is None:
            handle = self.current
        ast = self.visitor.ast_module(handle)
        self.log('Unparse ...')
        unparse.Unparser(ast, sys.stdout)
        print('\n')
        self.log('Done!', 'success')

    # Generator
    def generate(self, dst, dst_enc='utf-8'):
        # Concrete syntax tree
        output = os.path.join(dst, "cst.xml")
        self.log('Output %s' % output)
        tree = Tree.ElementTree(self.handles)
        tree.write(output, encoding=dst_enc)
        self.log('Done!', 'success')

        # Abstract syntax tree

        # Object dependency tree
        output = os.path.join(dst, "odt.xml")
        self.log('Output %s' % output)
        tree = Tree.ElementTree(self.xml)
        tree.write(output, encoding=dst_enc)
        self.log('Done!', 'success')

    @staticmethod
    def find_const(handles, value):
        h = handles.find("'" + value + "'")
        if h is not None:
            return h[3]


class Executor(object):
    operators = {
        '.': 'Add', '+': 'Add', '-': 'Sub', '*': 'Mult', '/': 'Div', '%': 'Mod',
        '**': 'Pow', '&&': 'And', '||': 'Or', '!': 'Not', '!==': 'NotEq', '<>': 'NotEq',
        '==': 'Eq', '===': 'Eq', '!=': 'NotEq', '<=': 'LtE', '<': 'Lt', '>': 'Gt', '>=': 'GtE',
        '>>': 'LShift', '<<': 'RShift', '|': 'BitOr', '&': 'BitAnd', '^': 'BitXor',}

    def __init__(self):
        self.tag_map = {}
        self.init_tag_map()

    def init_tag_map(self):
        pass

    def ast_module(self, handle):
        expr_ast = ast.parse('')
        for h in handle:
            stmt = self.ast_stmt(h)
            expr_ast.body.extend(stmt)
        exec compile(expr_ast, '<string>', 'exec')
        return expr_ast

    def ast_stmt(self, handle):
        stmt = []
        i = 0
        while len(handle) > i:
            name = handle[i].get('state')
            if name == 'label' or name == 'operator':
                block, count = self.ast_expr(handle, i)
                stmt.append(block)
                i += count
            if name == 'function':
                block, count = self.ast_function_def(handle, i)
                stmt.append(block)
                i += count
            else:
                i += 1
        return stmt
        # getattr(self, 'ast_%s' % name)(handle)

    def ast_class_def(self, handle, i):
        pass

    def ast_function_def(self, handle, i):
        name = handle[i+1].get('value').encode('utf-8')
        args = self.ast_argument(handle[i+2])
        body = self.ast_stmt(handle[i+3][1])
        fdef = ast.FunctionDef(name, args, body, [], lineno=5, col_offset=0)
        return fdef, 4

    def ast_if(self, handle, i):
        pass

    def ast_while(self, handle, i):
        pass

    def ast_for(self, handle, i):
        pass

    def ast_argument(self, handle):
        names = []
        defaults = []
        for h in handle:
            if h.get('state') in ('(', ')', ','):
                continue
            if h.get('state') == 'assign':
                var = h[0].get('value').strip('$').encode('utf-8')
                names.append(ast.Name(var, ast.Param(), lineno=5, col_offset=0))
                str = h[2][0].get('value').strip("'").encode('utf-8')
                defaults.append(ast.Str(str, lineno=5, col_offset=0))
            else:
                var = h[0].get('value').strip('$').encode('utf-8')
                names.append(ast.Name(var, ast.Param(), lineno=5, col_offset=0))
        args = ast.arguments(names, None, None, defaults, lineno=5, col_offset=0)
        return args

    def ast_expr(self, handle, i):
        count = 1
        n = handle[i].get('state')
        if n == 'label':
            label = handle[i].get('value').encode('utf-8')
            name = ast.Name(label, ast.Load(), lineno=5, col_offset=0)
            para = self.ast_parameter(handle[i+1])
            call = ast.Call(name, para, [], None, None, lineno=5, col_offset=0)
            expr = ast.Expr(call, lineno=5, col_offset=0)
            return expr, 2
        elif n == 'operator':
            str = handle[i+1].get('value').strip("'").encode('utf-8')
            left = ast.Str(str, lineno=5, col_offset=0)
            op = self.ast_operator(handle[i+2])
            label = handle[i+3].get('value').strip('$').encode('utf-8')
            right = ast.Name(label, ast.Load(), lineno=5, col_offset=0)
            bop = ast.BinOp(left, op, right, lineno=5, col_offset=0)
            expr = ast.Print(None, [bop], True, lineno=5, col_offset=0)
            return expr, 4

    def ast_parameter(self, handle):
        params = []
        for h in handle:
            if h.get('state') in ('(', ')', ','):
                continue
            str = h.get('value').strip("'").encode('utf-8')
            params.append(ast.Str(str, lineno=5, col_offset=0))
        return params

    def ast_operator(self, handle):
        if handle.get('value') in self.operators:
            op = self.operators[handle.get('value')]
            return eval("ast.%s()" % op)
        else:
            return None


class Visitor(Executor):

    def __init__(self, xml):
        Executor.__init__(self)

        self.xml = xml
        self.app = '.'
        self.root = '.'
        self.stack = []

    def set_path(self, root, app):
        self.root = root
        self.app = app

    def find_group(self, name):
        tags = name.split('.')
        root = self.xml
        for tag in tags:
            node = root.find(tag)
            if node is None:
                node = Tree.SubElement(root, tag)
            root = node
        return root

    def group_name(self, tag=None):
        names = []
        for i in self.stack:
            if i.group is None:
                continue
            names.append(i.group)
        if tag is None:
            return '.'.join(names)
        return '.'.join(names)+'.'+tag

    def push(self, handle):
        wrap = VisitorWrap(handle)
        self.stack.append(wrap)
        return wrap

    def pop(self):
        self.stack.pop()

    def visit(self, handle):
        if handle.get('state') in self.tag_map:
            wrap = self.push(handle)
            name = self.tag_map[handle.get('state')]
            getattr(self, 'visit_%s' % name)(handle, wrap)
        for h in handle:
            self.visit(h)
        else:
            if self.stack[-1].handle == handle:
                self.pop()

    def visit_file(self, handle, obj):

        sub = os.path.relpath(handle.get('value'), self.app)
        module = sub.split(os.path.sep)[0]
        filename = os.path.splitext(os.path.basename(handle.get('value')))[0]
        if module == '.' or module == '..':
            obj.group = 'global.' + filename
        else:
            obj.group = module + '.' + filename

        for h in handle:
            self.visit(h)

    def visit_namedblock(self, handle, obj):
        for h in handle:
            if h.get('state') == 'function':
                if obj.confirm > 0:
                    self.pop()
                    self.push(handle)
                obj.group = 'function'
                obj.confirm = 1
            elif h.get('state') == 'class':
                if obj.confirm > 0:
                    self.pop()
                    self.push(handle)
                obj.group = 'class'
                obj.confirm = 1
            elif h.get('state') == 'label':
                if obj.confirm < 2:
                    obj.group = h.get('value')
                    obj.confirm = 2
                    xml = self.find_group(self.group_name())

    def visit_statement(self, handle, obj):
        pass


class VisitorWrap(object):

    def __init__(self, handle):
        self.handle = handle
        self.state = handle.get('state')
        self.group = None
        self.confirm = 0
        self.attr = {}


class Generator(object):
    pass


class Handler(Tree.Element):

    def __init__(self, state, value=''):
        Tree.Element.__init__(self, 'node')
        self.set('state', state)
        self.set('value', value)
        self.content = value

    def add(self, val, operate):
        if operate == 'join':
            self.content += val.content
            self.set('value', self.get('value') + val.get('value'))
        else:
            if self.content != '':
                self.content += ' '
            self.content += val.content
            self.append(val)

    def find(self, value):
        for h in self:
            if h.get('value') == value:
                return self
            else:
                ret = h.find(value)
                if ret is not None:
                    return ret
        return None
