#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: windfic@qq.com
# @Last Modified Time: 2017-03-17 20:20

from compiler import Compiler
from compiler import Visitor
from compiler import Executor


class CompilerPhp(Compiler):

    def init_state_machine(self):
        self.state_machine = {
            'current': {
                'state': 'default', 'content': '', 'line': 0},
            'default': [
                {'name': 'open', 'next': 'php', 'extra': 0, 'start': 0, 'end': 0, 'cache': '',
                 'token': r'<\?'},
                {'name': 'open', 'next': 'php', 'extra': 0, 'start': 0, 'end': 0, 'cache': '',
                 'token': r'<\?php'}],
            'php': [
                {'name': 'close', 'next': 'default', 'extra': 0,
                 'token': r'\?>', 'start': 0, 'end': 0, 'cache': ''},
                {'name': 'lnum', 'next': '', 'extra': 0, 'start': 0, 'end': 0, 'cache': '',
                 'token': r'[0-9]+'},
                {'name': 'dnum', 'next': '', 'extra': 0, 'start': 0, 'end': 0, 'cache': '',
                 'token': r'([0-9]*\.[0-9]+)|([0-9]+\.[0-9]*)'},
                {'name': 'exponent', 'next': '', 'extra': 0, 'start': 0, 'end': 0, 'cache': '',
                 'token': r'(([0-9]+|([0-9]*\.[0-9]+)|([0-9]+\.[0-9]*))[eE][+-]?[0-9]+)'},
                {'name': 'hnum', 'next': '', 'extra': 0, 'start': 0, 'end': 0, 'cache': '',
                 'token': r'0x[0-9a-fA-F]+'},
                {'name': 'bnum', 'next': '', 'extra': 0, 'start': 0, 'end': 0, 'cache': '',
                 'token': r'0b[01]+'},
                {'name': 'label', 'next': '', 'extra': 0, 'start': 0, 'end': 0, 'cache': '',
                 'token': r'[a-zA-Z_\x7f-\xff][a-zA-Z0-9_\x7f-\xff]*'},
                {'name': 'var', 'next': '', 'extra': 0, 'start': 0, 'end': 0, 'cache': '',
                 'token': r'\$[a-zA-Z_\x7f-\xff][a-zA-Z0-9_\x7f-\xff]*'},
                {'name': 'comment', 'next': 'commentline', 'extra': 1,
                 'token': r'//', 'start': 0, 'end': 0, 'cache': ''},
                {'name': 'comment', 'next': 'commentline', 'extra': 1,
                 'token': r'#', 'start': 0, 'end': 0, 'cache': ''},
                {'name': 'comment', 'next': 'comment', 'extra': 1,
                 'token': r'/\*', 'start': 0, 'end': 0, 'cache': ''},
                {'name': 'string', 'next': 'string1', 'extra': 1,
                 'token': r'\'', 'start': 0, 'end': 0, 'cache': ''},
                {'name': 'string', 'next': 'string2', 'extra': 1,
                 'token': r'"', 'start': 0, 'end': 0, 'cache': ''},
                {'name': 'string', 'next': 'string3', 'extra': 1,
                 'token': r'<<<', 'start': 0, 'end': 0, 'cache': ''},
                {'name': 'symbol', 'next': '', 'extra': 0, 'start': 0, 'end': 0, 'cache': '',
                 'token': r'[\\\{\};:,\.\[\]\(\)\|\^&\+-/\*=%!~$<>\?@]'}],
            'string1': [
                {'name': 'string', 'next': 'php', 'extra': 0,
                 'token': r'\'', 'start': 0, 'end': 0, 'cache': ''},
                {'name': 'string', 'next': 'escape1', 'extra': 1,
                 'token': r'\\', 'start': 0, 'end': 0, 'cache': ''},
                {'name': 'string', 'next': '', 'extra': 1,
                 'token': r'', 'start': 0, 'end': 0, 'cache': ''}],
            'escape1': [
                {'name': 'string', 'next': 'string1', 'extra': 1,
                 'token': r'.', 'start': 0, 'end': 0, 'cache': ''}],
            'string2': [
                {'name': 'string', 'next': 'php', 'extra': 0,
                 'token': r'"', 'start': 0, 'end': 0, 'cache': ''},
                {'name': 'string', 'next': 'escape2', 'extra': 1,
                 'token': r'\\', 'start': 0, 'end': 0, 'cache': ''},
                {'name': 'string', 'next': '', 'extra': 1,
                 'token': r'', 'start': 0, 'end': 0, 'cache': ''}],
            'escape2': [
                {'name': 'string', 'next': 'string2', 'extra': 1,
                 'token': r'.', 'start': 0, 'end': 0, 'cache': ''}],
            'string3': [
                {'name': 'string', 'next': 'php', 'extra': 0,
                 'token': r'[dynamic]', 'start': 0, 'end': 0, 'cache': ''},
                {'name': 'string', 'next': '', 'extra': 1,
                 'token': r'', 'start': 0, 'end': 0, 'cache': ''}],
            'commentline': [
                {'name': 'comment', 'next': 'php', 'extra': 0,
                 'token': r'(\r|\n|\r\n)', 'start': 0, 'end': 0, 'cache': ''},
                {'name': 'comment', 'next': 'php', 'extra': 0,
                 'token': r'', 'start': 0, 'end': 0, 'cache': ''}],
            'comment': [
                {'name': 'comment', 'next': 'php', 'extra': 0,
                 'token': r'\*/', 'start': 0, 'end': 0, 'cache': ''},
                {'name': 'comment', 'next': '', 'extra': 1,
                 'token': r'', 'start': 0, 'end': 0, 'cache': ''}]
        }

    def init_action_table(self):
        self.action_table = {
            'stack': [],
            'extra': [],
            # accept
            '$$': {
                'close': {
                    'reduce': 'accept', 'op': 'ignore'},
                'statement': {
                    'reduce': 'accept', 'op': 'accept'},
                'namedblock': {
                    'reduce': 'accept', 'op': 'accept'}},
            'comment': {
                'reduce': 'accept', 'op': 'ignore'},
            'commentline': {
                'reduce': 'accept', 'op': 'ignore'},
            'open': {
                'reduce': 'accept', 'op': 'ignore'},
            'close': {
                'reduce': 'accept', 'op': 'ignore'},
            # class
            'namedblock': {
                'statement': {
                    'reduce': 'namedblock', 'op': 'merge'},
                'namedblock': {
                    'reduce': 'namedblock', 'op': 'merge'}},
            'class': {
                'statement': {  # anonymous class
                    'reduce': 'anonymous', 'op': 'part'},
                'label': {
                    'reduce': 'named', 'op': 'part'}},
            'named': {
                ';': {  # abstract
                    'reduce': 'namedblock', 'op': 'merge'},
                'statement': {
                    'reduce': 'namedblock', 'op': 'prefix'},
                'extends': {
                    'label': {  # class inheritance
                        'reduce': 'named', 'op': 'merge'}}},
            'interface': {
                'reduce': 'class', 'op': 'join'},
            'trait': {
                'reduce': 'class', 'op': 'join'},
            'new': {
                'var': {
                    'reduce': 'expression', 'op': 'part'},
                'label': {
                    'reduce': 'expression', 'op': 'part'},
                'anonymous': {  # anonymous class
                    'reduce': 'expression', 'op': 'part'}},
            'final': {
                'function': {
                    'reduce': 'function', 'op': 'part'}},
            'public': {
                'var': {
                    'reduce': 'var', 'op': 'part'},
                'function': {
                    'reduce': 'function', 'op': 'part'}},
            'protected': {
                'var': {
                    'reduce': 'var', 'op': 'part'},
                'function': {
                    'reduce': 'function', 'op': 'part'}},
            'private': {
                'var': {
                    'reduce': 'var', 'op': 'part'},
                'function': {
                    'reduce': 'function', 'op': 'part'}},
            'abstract': {
                'class': {
                    'reduce': 'class', 'op': 'part'},
                'function': {
                    'reduce': 'function', 'op': 'part'}},
            'static': {
                'var': {
                    'reduce': 'var', 'op': 'part'},
                'function': {
                    'reduce': 'function', 'op': 'part'}},
            'const': {
                'label': {
                    'reduce': 'var', 'op': 'merge'}},
            'namespace': {
                'label': {
                    ';': {
                        'reduce': 'statement', 'op': 'shift'}}},
            'use': {
                'label': {
                    ';': {
                        'reduce': 'statement', 'op': 'shift'}}},
            # function
            'function': {
                'parameter': {  # anonymous function
                    'reduce': 'nonamed', 'op': 'part'},
                'label': {
                    'parameter': {
                        'reduce': 'named', 'op': 'part'}}},
            'nonamed': {
                'statement': {
                    'reduce': 'anonymous', 'op': 'part'},
                'use': {
                    'parameter': {
                        'reduce': 'nonamed', 'op': 'part'}}},
            '(': {
                ')': {
                    'reduce': 'parameter', 'op': 'part'},
                'label': {
                    ')': {  # type
                        'reduce': 'parameter', 'op': 'part'}},
                'statement': {
                    ')': {
                        'reduce': 'parameter', 'op': 'part'}},
                'namedblock': {
                    ')': {
                        'reduce': 'parameter', 'op': 'part'}},
                'parameter': {
                    ')': {
                        'reduce': 'parameter', 'op': 'part'}},
                'expression': {
                    ')': {
                        'reduce': 'parameter', 'op': 'merge'}}},
            'label': {
                'parameter': {  # function call
                    'reduce': 'expression', 'op': 'part'},
                'extends': {
                    'label': {  # class inheritance
                        'reduce': 'label', 'op': 'part'}},
                'implements': {
                    'label': {
                        'reduce': 'label', 'op': 'part'}},
                ',': {
                    'expression': {
                        'reduce': 'expression', 'op': 'merge'}},
                '\\': {
                    'label': {
                        'reduce': 'label', 'op': 'part'}},
                'operator': {
                    'var': {  # $this->$var
                        'reduce': 'var', 'op': 'part'},
                    'label': {  # $this->CONST $this->function()
                        'reduce': 'label', 'op': 'part'},
                    'expression': {  # $this->$var
                        'reduce': 'expression', 'op': 'part'}}},
            '\\': {
                'label': {
                    'reduce': 'label', 'op': 'part',
                    '\\': {
                        'reduce': '\\', 'op': 'part'}}},
            # statement
            'goto': {
                'lable': {
                    ';': {
                        'reduce': 'statement', 'op': 'shift'}}},
            'break': {
                ';': {
                    'reduce': 'statement', 'op': 'shift'}},
            'continue': {
                ';': {
                    'reduce': 'statement', 'op': 'shift'}},
            'exit': {
                'reduce': 'expression', 'op': 'shift',
                'expression': {
                    'reduce': 'expression', 'op': 'part'}},
            'return': {
                ';': {
                    'reduce': 'statement', 'op': 'shift'},
                'expression': {
                    ';': {
                        'reduce': 'statement', 'op': 'shift'}}},
            'statement': {
                'namedblock': {
                    'reduce': 'namedblock', 'op': 'merge'},
                'statement': {
                    'reduce': 'statement', 'op': 'merge'}},
            '{': {
                '}': {
                    'reduce': 'statement', 'op': 'part'},
                'statement': {
                    '}': {
                        'reduce': 'statement', 'op': 'part'}},
                'namedblock': {
                    '}': {
                        'reduce': 'statement', 'op': 'part'}}},
            'condition': {
                ';': {
                    'reduce': 'statement', 'op': 'prefix'},
                'statement': {
                    'reduce': 'statement', 'op': 'prefix'}},
            # if
            'try': {
                'reduce': 'condition', 'op': 'part'},
            'catch': {
                'parameter': {
                    'reduce': 'condition', 'op': 'part'}},
            'if': {
                'parameter': {
                    'reduce': 'condition', 'op': 'part',
                    ':': {
                        'reduce': 'condition', 'op': 'part'}}},
            'elseif': {
                'parameter': {
                    'reduce': 'condition', 'op': 'part',
                    ':': {
                        'reduce': 'condition', 'op': 'part'}}},
            'else': {
                'reduce': 'condition', 'op': 'part',
                ':': {
                    'reduce': 'condition', 'op': 'part'}},
            'endif': {
                'reduce': 'condition'},
            # switch
            'switch': {
                'parameter': {
                    'reduce': 'condition', 'op': 'part'}},
            'case': {
                'expression': {
                    ':': {
                        'reduce': 'condition', 'op': 'part'}}},
            'default': {
                ':': {
                    'reduce': 'condition', 'op': 'part'}},
            # while
            'do': {
                'reduce': 'condition', 'op': 'part'},
            'while': {
                'parameter': {
                    'reduce': 'condition', 'op': 'part',
                    ':': {
                        'reduce': 'condition', 'op': 'part'}}},
            'endwhile': {
                'reduce': 'condition', 'op': 'part'},
            # for
            'for': {
                'parameter': {
                    'reduce': 'condition', 'op': 'part',
                    ':': {
                        'reduce': 'condition', 'op': 'part'}}},
            'endfor': {
                'reduce': 'condition', 'op': 'part'},
            'foreach': {
                'parameter': {
                    'reduce': 'condition', 'op': 'part'}},
            # expression
            'expression': {
                'postoperator': {
                    'reduce': 'expression', 'op': 'part'},
                'expression': {  # new Class or new Class()
                    'reduce': 'expression', 'op': 'part'},
                'parameter': {  # new Class or new Class()
                    'reduce': 'expression', 'op': 'prefix'},
                ';': {
                    'reduce': 'statement', 'op': 'merge'},
                ')': {  # for parameter
                    'reduce': ')', 'op': 'shift'},
                ',': {
                    'expression': {
                        'reduce': 'expression', 'op': 'merge'},
                    'label': {
                        'reduce': 'expression', 'op': 'merge'},
                    'anonymous': {
                        'reduce': 'expression', 'op': 'part'}},
                'operator': {
                    ')': {  # for parameter
                        'reduce': ')', 'op': 'shift'},
                    ';': {
                        'reduce': 'statement', 'op': 'shift'},
                    'anonymous': {
                        'reduce': 'expression', 'op': 'part'},
                    'label': {
                        'reduce': 'expression', 'op': 'part'},
                    'expression': {
                        'reduce': 'expression', 'op': 'merge'}}},
            'parameter': {
                'reduce': 'expression', 'op': 'part'},
            'var': {
                'reduce': 'expression', 'op': 'part',
                '=': {
                    'reduce': 'assign', 'op': 'part'}},
            'assign': {
                'label': {
                    'reduce': 'expression', 'op': 'part'},
                'anonymous': {
                    'reduce': 'expression', 'op': 'part'},
                'expression': {
                    'reduce': 'expression', 'op': 'append'}},
            ',': {
                ')': {  # 最后的逗号
                    'reduce': ')', 'op': 'select'},
                ']': {  # 最后的逗号
                    'reduce': ']', 'op': 'select'}},
            'operator': {
                'label': {
                    'reduce': 'expression', 'op': 'part'},
                'expression': {
                    'reduce': 'expression', 'op': 'merge'}},
            'null': {
                'reduce': 'expression', 'op': 'part'},
            'true': {
                'reduce': 'expression', 'op': 'part'},
            'false': {
                'reduce': 'expression', 'op': 'part'},
            'lnum': {
                'reduce': 'expression', 'op': 'part'},
            'dnum': {
                'reduce': 'expression', 'op': 'part'},
            'exponent': {
                'reduce': 'expression', 'op': 'part'},
            'bnum': {
                'reduce': 'expression', 'op': 'part'},
            'onum': {
                'reduce': 'expression', 'op': 'part'},
            'string': {
                'reduce': 'expression', 'op': 'part'},
            '?': {
                '?': {
                    'reduce': 'operator', 'op': 'join'},
                ':': {
                    'reduce': 'operator', 'op': 'join'},
                'label': {
                    ':': {
                        'reduce': 'operator', 'op': 'join'}},
                'expression': {
                    ':': {
                        'reduce': 'operator', 'op': 'part'}}},
            ':': {
                ':': {
                    'reduce': 'operator', 'op': 'join'}},
            '[': {
                ']': {
                    'reduce': 'operator', 'op': 'part'},
                'label': {
                    ']': {
                        'reduce': 'postoperator', 'op': 'part'}},
                'expression': {
                    ']': {
                        'reduce': 'postoperator', 'op': 'part'}}},
            # operator
            '>': {
                'reduce': 'operator', 'op': 'join',
                '=': {
                    'reduce': 'operator', 'op': 'join'},
                '>': {
                    'reduce': 'operator', 'op': 'join',
                    '=': {
                        'reduce': '=', 'op': 'join'}}},
            '<': {
                'reduce': 'operator', 'op': 'join',
                '>': {
                    'reduce': 'operator', 'op': 'join'},
                '<': {
                    'reduce': 'operator', 'op': 'join',
                    '=': {
                        'reduce': '=', 'op': 'join'}},
                '=': {
                    'reduce': 'operator', 'op': 'join',
                    '>': {
                        'reduce': 'operator', 'op': 'join'}}},
            '!': {
                'label': {
                    'reduce': 'expression', 'op': 'part'},
                'expression': {
                    'reduce': 'expression', 'op': 'part'},
                '=': {
                    'reduce': 'operator', 'op': 'join',
                    '=': {
                        'reduce': 'operator', 'op': 'join'}}},
            '=': {
                'reduce': 'operator', 'op': 'join',
                '>': {
                    'reduce': 'operator', 'op': 'join'},
                '=': {
                    'reduce': 'operator', 'op': 'join',
                    '=': {
                        'reduce': 'operator', 'op': 'join'}}},
            'and': {
                'reduce': 'operator', 'op': 'join'},
            'or': {
                'reduce': 'operator', 'op': 'join'},
            'xor': {
                'reduce': 'operator', 'op': 'join'},
            '&': {
                'reduce': 'operator', 'op': 'join',
                '&': {
                    'reduce': 'operator', 'op': 'join'},
                '=': {
                    'reduce': '=', 'op': 'join'}},
            '|': {
                'reduce': 'operator', 'op': 'join',
                '|': {
                    'reduce': 'operator', 'op': 'join'},
                '=': {
                    'reduce': '=', 'op': 'join'}},
            '+': {
                'reduce': 'operator', 'op': 'join',
                'expression': {
                    'reduce': 'expression', 'op': 'part'},
                '+': {
                    'reduce': 'operator', 'op': 'join'},
                '=': {
                    'reduce': '=', 'op': 'join'}},
            '-': {
                'reduce': 'operator', 'op': 'join',
                'expression': {
                    'reduce': 'expression', 'op': 'part'},
                '-': {
                    'reduce': 'operator', 'op': 'join'},
                '=': {
                    'reduce': '=', 'op': 'join'},
                '>': {
                    'reduce': 'operator', 'op': 'join'}},
            '*': {
                'reduce': 'operator', 'op': 'join',
                '=': {
                    'reduce': '=', 'op': 'join'},
                '*': {
                    'reduce': 'operator', 'op': 'join',
                    '=': {
                        'reduce': '=', 'op': 'join'}}},
            '/': {
                'reduce': 'operator', 'op': 'join',
                '=': {
                    'reduce': '=', 'op': 'join'}},
            '%': {
                'reduce': 'operator', 'op': 'join',
                '=': {
                    'reduce': '=', 'op': 'join'}},
            '.': {
                'reduce': 'operator', 'op': 'join',
                '=': {
                    'reduce': '=', 'op': 'join'}},
            '^': {
                'reduce': 'operator', 'op': 'join',
                '=': {
                    'reduce': '=', 'op': 'join'}},
            '~': {
                'reduce': 'operator', 'op': 'join'},
            '@': {
                'reduce': 'operator', 'op': 'join'},
            'as': {
                'reduce': 'operator', 'op': 'join'},
            'echo': {
                'reduce': 'operator', 'op': 'join'},
            'clone': {
                'reduce': 'operator', 'op': 'join'},
            'require': {
                'reduce': 'operator', 'op': 'join'},
            'throw': {
                'reduce': 'operator', 'op': 'join'},
            'include': {
                'reduce': 'operator', 'op': 'join'},
            'instanceof': {
                'reduce': 'operator', 'op': 'join'}
        }

    def init_keywords(self):
        self.keywords = [
            'abstract', 'and', 'as', 'break', 'callable', 'case', 'catch', 'clone',
            'const', 'class', 'continue', 'declare', 'default', 'do', 'echo', 'else',
            'elseif', 'enddeclare', 'endfor', 'endforeach', 'endif', 'endswitch', 'endwhile',
            'extends', 'final', 'finally', 'for', 'foreach', 'function', 'global', 'goto',
            'if', 'implements', 'include', 'include_once', 'instanceof', 'insteadof', 'interface',
            'namespace', 'new', 'or', 'print', 'private', 'protected', 'public', 'require',
            'return', 'static', 'switch', 'throw', 'trait', 'try', 'use', 'var',
            'while', 'xor', 'yield', 'true', 'false', 'null', 'exit', 'require']

    def init_visitor(self):
        self.visitor = VisitorPhp(self.xml)


class VisitorPhp(Visitor):

    def init_tag_map(self):
        self.tag_map = {
            'file': 'file', 'namedblock': 'namedblock', 'statement': 'statement',
        }


if __name__ == '__main__':
    CompilerPhp(1).compile_file('src/demo.php').execute()

