#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: windfic@qq.com
# @Last Modified Time: 2017-03-04 10:15

import os
import re
import chardet


class Processor(object):

    @staticmethod
    def detect(path):
        if not os.path.isfile(path):
            return 'auto'

        fh = open(path, 'r')
        try:
            data = fh.read()
        finally:
            fh.close()
        return chardet.detect(data)['encoding']

    @staticmethod
    def log(content, flag='info'):
        if flag=='info':
            print('>>> %s <<<'%content)
        elif flag=='success':
            print('>>> \033[1;32m%s :)\033[0m <<<'%content)
        else:
            print('>>> \033[1;31m%s :(\033[0m <<<'%content)


class Converter(Processor):
    separators = [('<', '>')]

    def __init__(self):
        self.match = {'opening': 0, 'closing': 0, 'extra': 0, 'script': '', 'content': '', 'cache': '', 'separator': ()}

    def convert_templates(self, src, dst, src_enc='auto', dst_enc='utf-8'):
        if not os.path.isdir(src) or not os.path.isdir(dst):
            return False

        for root, dirs, files in os.walk(src):
            for filepath in files:
                if src_enc == 'auto':
                    src_enc = self.detect(os.path.join(root, filepath))
                sub = os.path.relpath(root, src)
                self.convert_template(os.path.join(root, filepath), os.path.join(dst, sub, filepath), src_enc, dst_enc)

    def convert_template(self, src, dst, src_enc='auto', dst_enc='utf-8'):
        if not os.path.isfile(src):
            return False
        self.log(dst)
        if not os.path.exists(os.path.dirname(dst)):
            os.makedirs(os.path.dirname(dst))

        if src_enc == 'auto':
            src_enc = self.detect(src)

        sfh = open(src, 'r')
        dfh = open(dst, 'w')
        try:
            for line in sfh.readlines():
                line = line.decode(src_enc)
                for sep in self.separators:
                    line = self.next_tag(line, sep)
                    if line is None or line is '':
                        break
                if line is None or line is '':
                    continue
                line = line.encode(dst_enc)
                dfh.write(line)
            if self.match['content'] is not '' or self.match['cache'] is not '' or self.match['opening'] > 0:
                line = (self.match['content']+self.match['cache']).encode(dst_enc)
                self.match = {'opening': 0, 'closing': 0, 'extra': 0, 'script': '', 'content': '', 'cache': '', 'separator': ()}
                dfh.write(line)
        finally:
            sfh.close()
            dfh.close()

    def next_tag(self, line, sep):
        if self.match['opening'] > 0:
            n = line.find(self.match['separator'][1])
            if n >= 0:
                m = len(self.match['separator'][1])
                self.match['closing'] += 1
                self.match['opening'] += line[:n].count(self.match['separator'][0])
                self.match['cache'] += line[:n+m]
                if self.match['opening'] == self.match['closing']:
                    if len(line) > n + m:
                        conv = self.match['content']+self.convert_tag(self.match['cache'], line[n+m:])
                        m += self.match['extra']
                        self.match = {
                            'opening': 0, 'closing': 0, 'extra': 0, 'script': self.match['script'],
                            'content': conv, 'cache': '', 'separator': ()}
                        return self.next_tag(line[n+m:], sep)
                    else:
                        conv = self.match['content']+self.convert_tag(self.match['cache'])
                        self.match = {
                            'opening': 0, 'closing': 0, 'extra': 0, 'script': self.match['script'],
                            'content': '', 'cache': '', 'separator': ()}
                        return conv
                else:
                    if len(line) > n + m:
                        return self.next_tag(line[n + m:], sep)
            else:
                self.match['cache'] += line
        else:
            if self.match['script'] is not '':
                m = re.search(self.match['script'], line)
                if m:
                    self.match['script'] = ''
                    self.match['content'] += line[:m.end()]
                    if len(line) > m.end():
                        return self.next_tag(line[m.end():], sep)
                    else:
                        return self.next_tag('', sep)
                self.match['content'] += line
                return None
            else:
                n = line.find(sep[0])
                if n >= 0:
                    m = len(sep[0])
                    self.match['content'] += line[:n]
                    self.match['cache'] = line[n:n+m]
                    self.match['opening'] += 1
                    self.match['separator'] = sep
                    if len(line) > n + m:
                        return self.next_tag(line[n + m:], sep)

            content = self.match['content']+line
            self.match['content'] = ''
            return content
        return None

    def convert_tag(self, tag, extra=''):
        return tag


class ConverterThinkPhp2Jinja(Converter):
    separators = [('{', '}'), ('__', '__'), ('<', '>'), ('<?', '?>')]

    def convert_tag(self, tag, extra=''):
        m = re.match(r'__(\w+)__([^\'\"]*)', tag+extra)
        if m:
            return self.build_tag(tag, m.group(1).lower(), m.group(2))
        m = re.match(r'<\?(php)*(.*)\?>', tag)
        if m:
            return self.build_tag(m.group(2), '?php')
        m = re.match(r'<\s*(\w+)(\s+.*)*(/*)\s*>', tag)
        if m:
            return self.build_tag(tag, m.group(1).lower())
        m = re.match(r'<\s*/\s*(\w+)\s*>', tag)
        if m:
            return self.build_tag(tag, '/'+m.group(1).lower())
        m = re.match(r'\{\$\s*(.+)\s*\}', tag)
        if m:
            return self.build_tag(tag, '{$')
        m = re.match(r'\{:\s*(.+)\s*\}', tag)
        if m:
            return self.build_tag(tag, '{:')
        return tag

    def build_tag(self, tag, name, extra=''):
        if name == 'script':
            self.match['script'] = r'<\s*/\s*script\s*>'
        elif name == 'block':
            m = re.search(r'name\s*=\s*("|\')(.*?)("|\')', tag)
            if m:
                return u'{% block '+m.group(2)+' %}'
        elif name == '/block':
            return u'{% endblock %}'
        elif name == 'extend':
            m = re.search(r'name\s*=\s*("|\')(.*?)("|\')', tag)
            if m:
                return u'{% extends "'+m.group(2)+'.html" %}'
        elif name == '/extend':
            return u''
        elif name == 'include':
            m = re.search(r'file\s*=\s*("|\')(.*?)("|\')', tag)
            if m:
                return u'{% include "'+m.group(2)+'.html" %}'
        elif name == '/include':
            return u''
        elif name == 'volist':
            m1 = re.search(r'name\s*=\s*("(.*?)"|\'(.*?)\')', tag)
            m2 = re.search(r'id\s*=\s*("(.*?)"|\'(.*?)\')', tag)
            if m1 and m2:
                return u'{% for '+m2.group(1)[1:-1]+' in '+m1.group(1)[1:-1]+' %}'
        elif name == '/volist':
            return u'{% endfor %}'
        elif name == 'notempty':
            m = re.search(r'name\s*=\s*("(.*?)"|\'(.*?)\')', tag)
            if m:
                return u'{% if '+m.group(1)[1:-1]+' %}'
        elif name == '/notempty':
            return u'{% endif %}'
        elif name == '?php':
            return u'{# ' + tag + u' #}'
        elif name == 'php':
            return u'{# '
        elif name == '/php':
            return u' #}'
        elif name == '{$':
            m = re.search(r'\{\$\s*(.+)\s*\}', tag)
            if m:
                return u'{{ ' + m.group(1) + u' }}'
        elif name == '{:':
            m = re.search(r'\{:\s*(\w+)\s*\((.+)\)\s*\}', tag)
            if m:
                if m.group(1) == 'U':
                    return u'{{ url_for(' + m.group(2).replace('/', '.') + u') }}'
                return u'{{ ' + m.group(1) + '(' + m.group(2) + u') }}'
        elif name == 'static' or name == 'public':
            self.match['extra'] = len(extra)
            return u'{{ url_for(\'static\', filename=\'' + extra.lstrip('/') + '\') }}'
        return tag


class ConverterThinkPhp2Django(ConverterThinkPhp2Jinja):
    pass


if __name__ == '__main__':
    ConverterThinkPhp2Jinja().convert_templates('src', 'dst')
