#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: windfic@qq.com
# @Last Modified Time: 2017-03-17 20:36

import os
from converter import Processor
from compiler_php import CompilerPhp


class ConvertProject(Processor):

    def __init__(self):
        self.compiler = CompilerPhp()
        self.ext = '.php'

    def compile_project(self, src, dst, src_enc='auto', dst_enc='utf-8'):
        if not os.path.isdir(src) or not os.path.isdir(dst):
            return False

        index = os.path.join(src, 'index' + self.ext)
        if not os.path.exists(index):
            self.log('index%s is not found.' % self.ext, 'error')
            return False

        if src_enc == 'auto':
            src_enc = self.detect(index)

        self.compiler.compile_file(index, src_enc)
        app = self.compiler.find_const(self.compiler.handles, 'APP_PATH')
        if app is None:
            self.log('APP_PATH is not found in index%s!' % self.ext, 'error')
            return False

        app = os.path.join(src, app.get('value').strip("'"))
        if not os.path.exists(app):
            self.log(app + ' is not exists.', 'error')
        else:
            self.compiler.set_path(src, app)

        for root, dirs, files in os.walk(app):
            for filepath in files:
                path = os.path.join(root, filepath)
                if os.path.isfile(path):
                    if os.path.splitext(path)[1] != self.ext:
                        continue
                self.compiler.compile_file(path, src_enc)

        self.compiler.generate(dst, dst_enc)


if __name__ == '__main__':
    ConvertProject().compile_project('src', 'dst')

